@@ -12,14 +12,38 @@ use crate::clarity::contexts::{ContractContext, Environment, GlobalContext, Owne
 use crate::clarity::costs::cost_functions::ClarityCostFunction;
 use crate::clarity::database::{ClarityDatabase, NullBackingStore};
 use crate::clarity::errors::{Error, InterpreterResult};
-use crate::clarity::types::Value::UInt;
+use crate::clarity::types::{Value::UInt, PrincipalData, TupleData};
 use crate::clarity::types::{QualifiedContractIdentifier, TypeSignature, NONE};
 use crate::clarity::{ast, eval_all, ClarityName, SymbolicExpression, Value};
+use crate::clarity::types::signatures::TupleTypeSignature;
+use crate::clarity::types::{FunctionType, FunctionType::Fixed};
 
 type Result<T> = std::result::Result<T, CostErrors>;
 
 pub const CLARITY_MEMORY_LIMIT: u64 = 100 * 1000 * 1000;
 
+lazy_static! {
+    static ref COST_TUPLE_TYPE_SIGNATURE: TypeSignature = TypeSignature::TupleType(
+        TupleTypeSignature::try_from(vec![
+            ("runtime".into(), TypeSignature::UIntType),
+            ("write_length".into(), TypeSignature::UIntType),
+            ("write_count".into(), TypeSignature::UIntType),
+            ("read_count".into(), TypeSignature::UIntType),
+            ("read_length".into(), TypeSignature::UIntType),
+        ])
+        .expect("BUG: failed to construct type signature for cost tuple")
+    );
+}
+
+pub fn boot_code_id(contract_name: &str, is_mainnet: bool) -> QualifiedContractIdentifier {
+    let principal = if is_mainnet {
+        format!("SP000000000000000000002Q6VF78.{}", contract_name)
+    } else {
+        format!("ST000000000000000000002AMW42H.{}", contract_name)
+    };
+    QualifiedContractIdentifier::parse(&principal).unwrap()
+}
+
 pub fn runtime_cost<T: TryInto<u64>, C: CostTracker>(
     cost_function: ClarityCostFunction,
     tracker: &mut C,
@@ -191,8 +215,8 @@ pub struct LimitedCostTracker {
         HashMap<(QualifiedContractIdentifier, ClarityName), ClarityCostFunctionReference>,
     total: ExecutionCost,
     limit: ExecutionCost,
-    memory: u64,
-    memory_limit: u64,
+    pub memory: u64,
+    pub memory_limit: u64,
     free: bool,
     mainnet: bool,
 }
@@ -241,13 +265,270 @@ pub enum CostErrors {
     CostContractLoadFailure,
 }
 
+fn load_state_summary(mainnet: bool, clarity_db: &mut ClarityDatabase) -> Result<CostStateSummary> {
+    // let cost_voting_contract = boot_code_id("cost-voting", mainnet);
+
+    // let last_processed_at = match clarity_db.get_value(
+    //     "vm-costs::last-processed-at-height",
+    //     &TypeSignature::UIntType,
+    // ) {
+    //     Some(v) => u32::try_from(v.expect_u128()).expect("Block height overflowed u32"),
+    //     None => return Ok(CostStateSummary::empty()),
+    // };
+
+    // let metadata_result = clarity_db
+    //     .fetch_metadata_manual::<String>(
+    //         last_processed_at,
+    //         &cost_voting_contract,
+    //         "::state_summary",
+    //     )
+    //     .map_err(|e| CostErrors::CostComputationFailed(e.to_string()))?;
+    // let serialized: SerializedCostStateSummary = match metadata_result {
+    //     Some(serialized) => serde_json::from_str(&serialized).unwrap(),
+    //     None => return Ok(CostStateSummary::empty()),
+    // };
+    Ok(CostStateSummary::empty())
+}
+
+fn store_state_summary(
+    mainnet: bool,
+    clarity_db: &mut ClarityDatabase,
+    to_store: &CostStateSummary,
+) -> Result<()> {
+    let block_height = clarity_db.get_current_block_height();
+    let cost_voting_contract = boot_code_id("cost-voting", mainnet);
+
+    clarity_db.put(
+        "vm-costs::last-processed-at-height",
+        &Value::UInt(block_height as u128),
+    );
+    let serialized_summary =
+        serde_json::to_string(&SerializedCostStateSummary::from(to_store.clone()))
+            .expect("BUG: failure to serialize cost state summary struct");
+    clarity_db.set_metadata(
+        &cost_voting_contract,
+        "::state_summary",
+        &serialized_summary,
+    );
+
+    Ok(())
+}
+
+///
+/// This method loads a cost state summary structure from the currently open stacks chain tip
+///   In doing so, it reads from the cost-voting contract to find any newly confirmed proposals,
+///    checks those proposals for validity, and then applies those changes to the cached set
+///    of cost functions.
+///
+/// `apply_updates` - tells this function to look for any changes in the cost voting contract
+///   which would need to be applied. if `false`, just load the last computed cost state in this
+///   fork.
+///
+fn load_cost_functions(
+    mainnet: bool,
+    clarity_db: &mut ClarityDatabase,
+    apply_updates: bool,
+) -> Result<CostStateSummary> {
+    if mainnet == false {
+        return Ok(CostStateSummary::empty())
+    }
+
+    let last_processed_count = clarity_db
+        .get_value("vm-costs::last_processed_count", &TypeSignature::UIntType)
+        .unwrap_or(Value::UInt(0))
+        .expect_u128();
+    let cost_voting_contract = boot_code_id("cost-voting", mainnet);
+    let confirmed_proposals_count = clarity_db
+        .lookup_variable_unknown_descriptor(&cost_voting_contract, "confirmed-proposal-count")
+        .map_err(|e| CostErrors::CostComputationFailed(e.to_string()))?
+        .expect_u128();
+    println!("Check cost voting contract");
+
+    // we need to process any confirmed proposals in the range [fetch-start, fetch-end)
+    let (fetch_start, fetch_end) = (last_processed_count, confirmed_proposals_count);
+    let mut state_summary = load_state_summary(mainnet, clarity_db)?;
+    if !apply_updates {
+        return Ok(state_summary);
+    }
+
+    for confirmed_proposal in fetch_start..fetch_end {
+        // fetch the proposal data
+        let entry = clarity_db
+            .fetch_entry_unknown_descriptor(
+                &cost_voting_contract,
+                "confirmed-proposals",
+                &Value::from(
+                    TupleData::from_data(vec![(
+                        "confirmed-id".into(),
+                        Value::UInt(confirmed_proposal),
+                    )])
+                    .expect("BUG: failed to construct simple tuple"),
+                ),
+            )
+            .expect("BUG: Failed querying confirmed-proposals")
+            .expect_optional()
+            .expect("BUG: confirmed-proposal-count exceeds stored proposals")
+            .expect_tuple();
+        let target_contract = match entry
+            .get("function-contract")
+            .expect("BUG: malformed cost proposal tuple")
+            .clone()
+            .expect_principal()
+        {
+            PrincipalData::Contract(contract_id) => contract_id,
+            _ => {
+                println!("Confirmed cost proposal invalid: function-contract is not a contract principal");
+                continue;
+            }
+        };
+        let target_function = match ClarityName::try_from(
+            entry
+                .get("function-name")
+                .expect("BUG: malformed cost proposal tuple")
+                .clone()
+                .expect_ascii(),
+        ) {
+            Ok(x) => x,
+            Err(_) => {
+                println!("Confirmed cost proposal invalid: function-name is not a valid function name");
+                continue;
+            }
+        };
+        let cost_contract = match entry
+            .get("cost-function-contract")
+            .expect("BUG: malformed cost proposal tuple")
+            .clone()
+            .expect_principal()
+        {
+            PrincipalData::Contract(contract_id) => contract_id,
+            _ => {
+                println!("Confirmed cost proposal invalid: cost-function-contract is not a contract principal");
+                continue;
+            }
+        };
+
+        let cost_function = match ClarityName::try_from(
+            entry
+                .get_owned("cost-function-name")
+                .expect("BUG: malformed cost proposal tuple")
+                .expect_ascii(),
+        ) {
+            Ok(x) => x,
+            Err(_) => {
+                println!("Confirmed cost proposal invalid: cost-function-name is not a valid function name");
+                continue;
+            }
+        };
+
+        // Here is where we perform the required validity checks for a confirmed proposal:
+        //  * Replaced contract-calls _must_ be `define-read-only` _or_ refer to one of the boot code
+        //      cost functions
+        //  * cost-function contracts must be arithmetic only
+
+        // make sure the contract is "cost contract eligible" via the
+        //  arithmetic-checking analysis pass
+        let (cost_func_ref, cost_func_type) = match clarity_db
+            .load_contract_analysis(&cost_contract)
+        {
+            Some(c) => {
+                if !c.is_cost_contract_eligible {
+                    println!("Confirmed cost proposal invalid: cost-function-contract uses non-arithmetic or otherwise illegal operations");
+                    continue;
+                }
+
+                if let Some(FunctionType::Fixed(cost_function_type)) = c
+                    .read_only_function_types
+                    .get(&cost_function)
+                    .or_else(|| c.private_function_types.get(&cost_function))
+                {
+                    if !cost_function_type.returns.eq(&COST_TUPLE_TYPE_SIGNATURE) {
+                        println!("Confirmed cost proposal invalid: cost-function-name does not return a cost tuple");
+                        continue;
+                    }
+                    if !cost_function_type.args.len() == 1
+                        || cost_function_type.args[0].signature != TypeSignature::UIntType
+                    {
+                        println!("Confirmed cost proposal invalid: cost-function-name args should be length-1 and only uint");
+                        continue;
+                    }
+                    (
+                        ClarityCostFunctionReference {
+                            contract_id: cost_contract,
+                            function_name: cost_function.to_string(),
+                        },
+                        cost_function_type.clone(),
+                    )
+                } else {
+                    println!("Confirmed cost proposal invalid: cost-function-name not defined");
+                    continue;
+                }
+            }
+            None => {
+                println!("Confirmed cost proposal invalid: cost-function-contract is not a published contract");
+                continue;
+            }
+        };
+
+        if target_contract == boot_code_id("costs", mainnet) {
+            // refering to one of the boot code cost functions
+            let target = match ClarityCostFunction::lookup_by_name(&target_function) {
+                Some(cost_func) => cost_func,
+                None => {
+                    println!("Confirmed cost proposal invalid: function-name does not reference a Clarity cost function");
+                    continue;
+                }
+            };
+            state_summary
+                .cost_function_references
+                .insert(target, cost_func_ref);
+        } else {
+            // referring to a user-defined function
+            match clarity_db.load_contract_analysis(&target_contract) {
+                Some(c) => {
+                    if let Some(Fixed(tf)) = c.read_only_function_types.get(&target_function) {
+                        if cost_func_type.args.len() != tf.args.len() {
+                            println!("Confirmed cost proposal invalid: cost-function contains the wrong number of arguments");
+                            continue;
+                        }
+                        for arg in &cost_func_type.args {
+                            if &arg.signature != &TypeSignature::UIntType {
+                                println!("Confirmed cost proposal invalid: contains non uint argument");
+                                continue;
+                            }
+                        }
+                    } else {
+                        println!("Confirmed cost proposal invalid: function-name not defined or is not read-only");
+                        continue;
+                    }
+                }
+                None => {
+                    println!("Confirmed cost proposal invalid: contract-name not a published contract");
+                    continue;
+                }
+            }
+            state_summary
+                .contract_call_circuits
+                .insert((target_contract, target_function), cost_func_ref);
+        }
+    }
+    if confirmed_proposals_count > last_processed_count {
+        store_state_summary(mainnet, clarity_db, &state_summary)?;
+        clarity_db.put(
+            "vm-costs::last_processed_count",
+            &Value::UInt(confirmed_proposals_count),
+        );
+    }
+
+    Ok(state_summary)
+}
+
 impl LimitedCostTracker {
     pub fn new(
         mainnet: bool,
         limit: ExecutionCost,
         clarity_db: &mut ClarityDatabase,
     ) -> Result<LimitedCostTracker> {
-        let cost_tracker = LimitedCostTracker {
+        let mut cost_tracker = LimitedCostTracker {
             cost_function_references: HashMap::new(),
             cost_contracts: HashMap::new(),
             contract_call_circuits: HashMap::new(),
@@ -258,7 +539,7 @@ impl LimitedCostTracker {
             free: false,
             mainnet,
         };
-        // cost_tracker.load_costs(clarity_db, true)?;
+        cost_tracker.load_costs(clarity_db, true)?;
         Ok(cost_tracker)
     }
 
@@ -267,7 +548,7 @@ impl LimitedCostTracker {
         limit: ExecutionCost,
         clarity_db: &mut ClarityDatabase,
     ) -> Result<LimitedCostTracker> {
-        let cost_tracker = LimitedCostTracker {
+        let mut cost_tracker = LimitedCostTracker {
             cost_function_references: HashMap::new(),
             cost_contracts: HashMap::new(),
             contract_call_circuits: HashMap::new(),
@@ -278,7 +559,7 @@ impl LimitedCostTracker {
             free: false,
             mainnet,
         };
-        // cost_tracker.load_costs(clarity_db, false)?;
+        cost_tracker.load_costs(clarity_db, false)?;
         Ok(cost_tracker)
     }
 
@@ -295,47 +576,68 @@ impl LimitedCostTracker {
             mainnet: false,
         }
     }
-    pub fn load_boot_costs(&mut self, clarity_db: &mut ClarityDatabase) -> Result<()> {
-        // let boot_costs_id = (*STACKS_BOOT_COST_CONTRACT).clone();
 
-        // clarity_db.begin();
+    /// `apply_updates` - tells this function to look for any changes in the cost voting contract
+    ///   which would need to be applied. if `false`, just load the last computed cost state in this
+    ///   fork.
+    fn load_costs(&mut self, clarity_db: &mut ClarityDatabase, apply_updates: bool) -> Result<()> {
+        let boot_costs_id = boot_code_id("costs", self.mainnet);
 
-        // let mut cost_contracts = HashMap::new();
-        // let mut m = HashMap::new();
-        // for f in ClarityCostFunction::ALL.iter() {
-        //     m.insert(
-        //         f,
-        //         ClarityCostFunctionReference::new(boot_costs_id.clone(), f.get_name()),
-        //     );
-        //     if !cost_contracts.contains_key(&boot_costs_id) {
-        //         let contract_context = match clarity_db.get_contract(&boot_costs_id) {
-        //             Ok(contract) => contract.contract_context,
-        //             Err(e) => {
-        //                 clarity_db.roll_back();
-        //                 return Err(CostErrors::CostContractLoadFailure);
-        //             }
-        //         };
-        //         cost_contracts.insert(boot_costs_id.clone(), contract_context);
-        //     }
-        // }
+        clarity_db.begin();
+        let CostStateSummary {
+            contract_call_circuits,
+            mut cost_function_references,
+        } = load_cost_functions(self.mainnet, clarity_db, apply_updates).map_err(|e| {
+            clarity_db.roll_back();
+            e
+        })?;
 
-        // for (_, circuit_target) in self.contract_call_circuits.iter() {
-        //     if !cost_contracts.contains_key(&circuit_target.contract_id) {
-        //         let contract_context = match clarity_db.get_contract(&circuit_target.contract_id) {
-        //             Ok(contract) => contract.contract_context,
-        //             Err(e) => {
-        //                 clarity_db.roll_back();
-        //                 return Err(CostErrors::CostContractLoadFailure);
-        //             }
-        //         };
-        //         cost_contracts.insert(circuit_target.contract_id.clone(), contract_context);
-        //     }
-        // }
+        self.contract_call_circuits = contract_call_circuits;
 
-        // self.cost_function_references = m;
-        // self.cost_contracts = cost_contracts;
+        let mut cost_contracts = HashMap::new();
+        let mut m = HashMap::new();
+        for f in ClarityCostFunction::ALL.iter() {
+            let cost_function_ref = cost_function_references.remove(&f).unwrap_or_else(|| {
+                ClarityCostFunctionReference::new(boot_costs_id.clone(), f.get_name())
+            });
+            if !cost_contracts.contains_key(&cost_function_ref.contract_id) {
+                let contract_context = match clarity_db.get_contract(&cost_function_ref.contract_id)
+                {
+                    Ok(contract) => contract.contract_context,
+                    Err(e) => {
+                        println!("Failed to load intended Clarity cost contract");
+                        clarity_db.roll_back();
+                        return Err(CostErrors::CostContractLoadFailure);
+                    }
+                };
+                cost_contracts.insert(cost_function_ref.contract_id.clone(), contract_context);
+            }
 
-        // clarity_db.roll_back();
+            m.insert(f, cost_function_ref);
+        }
+
+        for (_, circuit_target) in self.contract_call_circuits.iter() {
+            if !cost_contracts.contains_key(&circuit_target.contract_id) {
+                let contract_context = match clarity_db.get_contract(&circuit_target.contract_id) {
+                    Ok(contract) => contract.contract_context,
+                    Err(e) => {
+                        println!("Failed to load intended Clarity cost contract");
+                        clarity_db.roll_back();
+                        return Err(CostErrors::CostContractLoadFailure);
+                    }
+                };
+                cost_contracts.insert(circuit_target.contract_id.clone(), contract_context);
+            }
+        }
+
+        self.cost_function_references = m;
+        self.cost_contracts = cost_contracts;
+
+        if apply_updates {
+            clarity_db.commit();
+        } else {
+            clarity_db.roll_back();
+        }
 
         return Ok(());
     }
@@ -555,7 +857,7 @@ impl CostTracker for &mut LimitedCostTracker {
     }
 }
 
-#[derive(Debug, Deserialize, Serialize, Clone, PartialEq, Eq)]
+#[derive(Debug, Default, Deserialize, Serialize, Clone, PartialEq, Eq)]
 pub struct ExecutionCost {
     pub write_length: u64,
     pub write_count: u64,
@@ -603,6 +905,21 @@ impl ExecutionCost {
         }
     }
 
+    /// Returns the percentage of self consumed in `numerator`'s largest proportion dimension.
+    pub fn proportion_largest_dimension(&self, numerator: &ExecutionCost) -> u64 {
+        [
+            numerator.runtime / cmp::max(1, self.runtime / 100),
+            numerator.write_length / cmp::max(1, self.write_length / 100),
+            numerator.write_count / cmp::max(1, self.write_count / 100),
+            numerator.read_length / cmp::max(1, self.read_length / 100),
+            numerator.read_count / cmp::max(1, self.read_count / 100),
+        ]
+        .iter()
+        .max()
+        .expect("BUG: should find maximum")
+        .clone()
+    }
+
     pub fn max_value() -> ExecutionCost {
         Self {
             runtime: u64::max_value(),
